/*
 * LSM_ReaderM.java
 *
 * @author Patrick Pirrotte
 *
 * ImageJ plugin to read Zeiss LSM images
 *
 * !!! The complete plugin is available at the Imagej plugin page, with support for LSM
 * !!! extended tags, events, lambda and time stamps and much more...
 *
 * Version  :      3.2g minimal
 * 
 * Authors  :      Patrick PIRROTTE (patrick.pirrotte@gmx.net)
 *                 Jerome Mutterer (jerome.mutterer@ibmp-ulp.u-strasbg.fr)
 *                 Yannick KREMPP (till v2.2) (yannick.krempp@iphysiol.unil.ch)
 *
 *                 written for the IBMP-CNRS Strasbourg(France)
 *                 
 * Description :   This plugin reads *.lsm files produced by Zeiss LSM 510 confocal microscopes.
 *                 Each channel of an image plane, stack or time series is opened as a separate
 *                 image or stack window. The plugin also retrieves calibration infos from
 *                 LSM files. This plugin has been devlopped using Zeiss' v3.2 fileformat
 *                 specifications. This software is compatible with files generated with AIM version 3.2.
 *                 Other versions of the lsm format should be readable more or less well.
 *                 A short manual is available from
 *                 ibmp.u-strasbg.fr/sg/microscopie/methods/lsmio/lsmio.htm
 *
 */
import ij.plugin.PlugIn;
import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.Menus;
import ij.io.FileInfo;
import ij.io.FileOpener;
import ij.io.ImageReader;
import ij.io.OpenDialog;
import ij.io.PluginClassLoader;
import ij.io.RandomAccessStream;
import ij.measure.Calibration;
import ij.process.ImageProcessor;

import java.awt.Color;
import java.awt.image.ColorModel;
import java.awt.image.IndexColorModel;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.Vector;

public class LSM_Reader implements PlugIn {

    public void run(String arg) {
		if (arg != "") {
			File temp = new File(arg);
			String arg2 = temp.getName();
			String arg1 = "";
			arg1 = temp.getPath();
			arg1 = arg1.substring(0, (arg1.length() - arg2.length()));
			openLSM(arg1, arg2);
		} else
			openLSM(arg);
    } /* openLSM, opens the LSM file */

    public ImagePlus[] openLSM(String arg) {
        String directory, filename;
        OpenDialog od = new OpenDialog("Open LSM image ... ", arg);
        filename = od.getFileName();

        if (filename == null) {
            IJ.error("no file selected");
            return null;
        }
        directory = od.getDirectory();
        if (directory == null) {
            IJ.error("no file selected");
            return null;
        }
        ImagePlus[] imp = openLSM(directory, filename);
        if (imp != null) {
            for (int i = 0; i < imp.length; i++) {

                imp[i].show();
                imp[i].updateAndDraw();
            }
        }
        return imp;
    }

    public ImagePlus[] openLSM(String directory, String filename) {
        ImagePlus[] imp = null;
        RandomAccessFile file;
        LsmFileInfo lsm;
        try {
            file = new RandomAccessFile(new File(directory, filename), "r");
            RandomAccessStream stream = new RandomAccessStream(file);
            lsm = new LsmFileInfo();
            lsm.fileName = filename;
            lsm.directory = directory;
            if (isLSMfile(stream)) {
                ImageDirectory imDir = readImageDirectoy(stream, 8);
                lsm.imageDirectories.add(imDir);
                while (imDir.OFFSET_NEXT_DIRECTORY != 0) {
                    imDir = readImageDirectoy(stream,
                            imDir.OFFSET_NEXT_DIRECTORY);
                    lsm.imageDirectories.add(imDir);
                }
                imp = open(stream, lsm);
                stream.close();

            } else
                IJ.error("Not a valid lsm file");
        } catch (FileNotFoundException e) {
            IJ.error("File not found");
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
        return imp;
    }

    private boolean isLSMfile(RandomAccessStream stream) {
        boolean identifier = false;
        long ID = 0;
        try {
            stream.seek(2);
            ID = swap(stream.readShort());
            if (ID == 42)
                identifier = true;
        } catch (IOException e) {
            e.printStackTrace();
        }
        return identifier;
    }

    private long getTagCount(RandomAccessStream stream, long position) {
        long tags = 0;
        try {
            stream.seek((int) position);
            tags = swap(stream.readShort());
        } catch (IOException e) {
            e.printStackTrace();
        }
        return tags;
    }

    private ImageDirectory readImageDirectoy(RandomAccessStream stream,
            long startPosition) {
        ImageDirectory imDir = new ImageDirectory();
        long tags = getTagCount(stream, startPosition);
        byte[] tag;

        int tagtype = 0;
        int MASK = 0x00ff;
        long MASK2 = 0x000000ff;
        long currentTagPosition = 0;
        long stripOffset = 0, stripByteOffset = 0;

        for (int i = 0; i < tags; i++) {
            currentTagPosition = startPosition + 2 + i * 12;
            tag = readTag(stream, (int) currentTagPosition);
            tagtype = ((tag[1] & MASK) << 8) | ((tag[0] & MASK) << 0);

            switch (tagtype) {
            case 254:
                imDir.TIF_NEWSUBFILETYPE = ((tag[11] & MASK2) << 24)
                        | ((tag[10] & MASK2) << 16) | ((tag[9] & MASK2) << 8)
                        | (tag[8] & MASK2);
                break;
            case 256:
                imDir.TIF_IMAGEWIDTH = ((tag[11] & MASK2) << 24)
                        | ((tag[10] & MASK2) << 16) | ((tag[9] & MASK2) << 8)
                        | (tag[8] & MASK2);
                break;
            case 257:
                imDir.TIF_IMAGELENGTH = ((tag[11] & MASK2) << 24)
                        | ((tag[10] & MASK2) << 16) | ((tag[9] & MASK2) << 8)
                        | (tag[8] & MASK2);
                break;
            case 258:
                imDir.TIF_BITSPERSAMPLE_LENGTH = ((tag[7] & MASK2) << 24)
                        | ((tag[6] & MASK2) << 16) | ((tag[5] & MASK2) << 8)
                        | (tag[4] & MASK2);
                imDir.TIF_BITSPERSAMPLE_CHANNEL[0] = ((tag[8] & MASK2) << 0);
                imDir.TIF_BITSPERSAMPLE_CHANNEL[1] = ((tag[9] & MASK2) << 0);
                imDir.TIF_BITSPERSAMPLE_CHANNEL[2] = ((tag[10] & MASK2) << 0);
                break;
            case 259:
                imDir.TIF_COMPRESSION = ((tag[8] & MASK2) << 0);
                break;
            case 262:
                imDir.TIF_PHOTOMETRICINTERPRETATION = ((tag[8] & MASK2) << 0);
                break;
            case 273:
                imDir.TIF_STRIPOFFSETS_LENGTH = ((tag[7] & MASK2) << 24)
                        | ((tag[6] & MASK2) << 16) | ((tag[5] & MASK2) << 8)
                        | (tag[4] & MASK2);
                stripOffset = ((tag[11] & MASK2) << 24)
                        | ((tag[10] & MASK2) << 16) | ((tag[9] & MASK2) << 8)
                        | (tag[8] & MASK2);
                break;
            case 277:
                imDir.TIF_SAMPLESPERPIXEL = ((tag[8] & MASK2) << 0);
                break;
            case 279:
                imDir.TIF_STRIPBYTECOUNTS_LENGTH = ((tag[7] & MASK2) << 24)
                        | ((tag[6] & MASK2) << 16) | ((tag[5] & MASK2) << 8)
                        | (tag[4] & MASK2);
                stripByteOffset = ((tag[11] & MASK2) << 24)
                        | ((tag[10] & MASK2) << 16) | ((tag[9] & MASK2) << 8)
                        | (tag[8] & MASK2);
                break;
            case 317:
                imDir.TIF_PREDICTOR = ((tag[8] & MASK2) << 0);
                break;
            case 34412:
                imDir.TIF_CZ_LSMINFO_OFFSET = ((tag[11] & MASK2) << 24)
                        | ((tag[10] & MASK2) << 16) | ((tag[9] & MASK2) << 8)
                        | (tag[8] & MASK2);
                break;
            default:
                break;
            }
        }
        imDir.TIF_STRIPOFFSETS = new long[(int) imDir.TIF_STRIPOFFSETS_LENGTH];
        if (imDir.TIF_STRIPOFFSETS_LENGTH == 1)
            imDir.TIF_STRIPOFFSETS[0] = stripOffset;
        else
            imDir.TIF_STRIPOFFSETS = getIntTable(stream, stripOffset,
                    (int) imDir.TIF_STRIPOFFSETS_LENGTH);

        imDir.TIF_STRIPBYTECOUNTS = new long[(int) imDir.TIF_STRIPBYTECOUNTS_LENGTH];
        imDir.TIF_STRIPBYTECOUNTS = getIntTable(stream, stripByteOffset,
                (int) imDir.TIF_STRIPBYTECOUNTS_LENGTH);

        try {
            stream.seek((int) (currentTagPosition + 12));
            int offset_next_directory = swap(stream.readInt());
            imDir.OFFSET_NEXT_DIRECTORY = offset_next_directory;
        } catch (IOException e) {
            e.printStackTrace();
        }
        if (imDir.TIF_CZ_LSMINFO_OFFSET != 0) {
            imDir.TIF_CZ_LSMINFO = getCZ_LSMINFO(stream,
                    imDir.TIF_CZ_LSMINFO_OFFSET);
        }
        return imDir;
    }

    private byte[] readTag(RandomAccessStream stream, int position) {
        byte[] tag = new byte[12];
        try {
            stream.seek(position);
            stream.readFully(tag);
        } catch (IOException e) {
            e.printStackTrace();
        }
        return tag;
    }

    private long[] getIntTable(RandomAccessStream stream, long position,
            int count) {
        long[] offsets = new long[count];
        try {
            stream.seek((int) position);
            for (int i = 0; i < count; i++)
                offsets[i] = swap(stream.readInt());
        } catch (IOException e) {
            e.printStackTrace();
        }
        return offsets;
    }

    private CZ_LSMInfo getCZ_LSMINFO(RandomAccessStream stream, long position) {
        CZ_LSMInfo cz = new CZ_LSMInfo();
        try {
            if (position == 0)
                return cz;
            stream.seek((int) position + 20);
            cz.DimensionChannels = swap(stream.readInt());
            stream.seek((int) position + 28);
            cz.IntensityDataType = swap(stream.readInt());

            stream.seek((int) position + 40);
            cz.VoxelSizeX = swap(stream.readDouble());
            cz.VoxelSizeY = swap(stream.readDouble());
            cz.VoxelSizeZ = swap(stream.readDouble());
            stream.seek((int) position + 88);
            cz.ScanType = swap(stream.readShort());
            stream.seek((int) position + 108);
            cz.OffsetChannelColors = swap(stream.readInt());
            if (cz.OffsetChannelColors != 0) {
                ChannelNamesAndColors channelNamesAndColors = getChannelNamesAndColors(
                        stream, cz.OffsetChannelColors, cz.DimensionChannels);
                cz.channelNamesAndColors = channelNamesAndColors;
            }

        } catch (IOException getCZ_LSMINFO_exception) {
            getCZ_LSMINFO_exception.printStackTrace();
        }

        return cz;
    }

    private ChannelNamesAndColors getChannelNamesAndColors(
            RandomAccessStream stream, long position, long channelCount) {
        ChannelNamesAndColors channelNamesAndColors = new ChannelNamesAndColors();
        try {
            stream.seek((int) position);
            channelNamesAndColors.BlockSize = swap(stream.readInt());
            channelNamesAndColors.NumberColors = swap(stream.readInt());
            channelNamesAndColors.NumberNames = swap(stream.readInt());
            channelNamesAndColors.ColorsOffset = swap(stream.readInt());
            channelNamesAndColors.NamesOffset = swap(stream.readInt());
            channelNamesAndColors.Mono = swap(stream.readInt()); // reserved
            // 4 words
            stream.seek((int) channelNamesAndColors.NamesOffset
                    + (int) position);
            channelNamesAndColors.ChannelNames = new String[(int) channelCount];
            long Namesize = channelNamesAndColors.BlockSize
                    - channelNamesAndColors.NamesOffset;
            byte[] buffer = new byte[(int) Namesize];
            stream.read(buffer);
            String allNames = new String(buffer);

            int begindex = 4;
            int endindex = 5;
            for (int j = 0; j < channelCount; j++) {
                endindex = allNames.indexOf(00, begindex);
                channelNamesAndColors.ChannelNames[j] = allNames.substring(
                        begindex, endindex);
                begindex = endindex + 5;
            }

            stream.seek((int) channelNamesAndColors.ColorsOffset
                    + (int) position);
            channelNamesAndColors.Colors = new long[(int) (channelNamesAndColors.NumberColors)];

            for (int j = 0; j < (int) (channelNamesAndColors.NumberColors); j++) {
                channelNamesAndColors.Colors[j] = swap(stream.readInt());
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
        return channelNamesAndColors;
    }

    public ImagePlus[] open(RandomAccessStream stream, LsmFileInfo lsmFi) {
        ImageDirectory firstImDir = (ImageDirectory) lsmFi.imageDirectories
                .get(0);
        if (firstImDir == null) {
            IJ.error("LSM ImageDir null.");
            return null;
        }
        CZ_LSMInfo cz = firstImDir.TIF_CZ_LSMINFO;
        if (cz == null) {
            IJ.error("LSM ImageDir null (CZ missing).");
            return null;
        }
        if (firstImDir.TIF_COMPRESSION != 1) {
            IJ.error("Compressed LSM images are not supported yet.");
            return null;
        }
        FileOpener fop = new FileOpener(lsmFi);
        ColorModel cm = fop.createColorModel(lsmFi);
        lsmFi = populateImageSpace(lsmFi, cz, firstImDir.TIF_COMPRESSION,
                firstImDir.TIF_PREDICTOR, firstImDir.TIF_STRIPOFFSETS,
                firstImDir.TIF_STRIPBYTECOUNTS);
        ImagePlus[] imps = null;
        switch (cz.ScanType) {
        case 0:
            imps = readStack(stream, lsmFi, cz, cm, fop);
            return imps;
        case 1:
            imps = readStack(stream, lsmFi, cz, cm, fop);
            return imps;
        case 3:
            imps = readStack(stream, lsmFi, cz, cm, fop);
            return imps;
        case 4:
            imps = readStack(stream, lsmFi, cz, cm, fop);
            return imps;
        case 5:
            imps = readStack(stream, lsmFi, cz, cm, fop);
            return imps;
        case 6:
            imps = readStack(stream, lsmFi, cz, cm, fop);
            return imps;
        default:
            IJ.error("Unsupported LSM scantype: " + cz.ScanType);
            break;
        }
        return imps;
    }

    private LsmFileInfo populateImageSpace(LsmFileInfo lsmFi, CZ_LSMInfo cz,
            long compression, long predictor, long[] stripOffsets,
            long[] stripLengths) {
        lsmFi.url = "";
        lsmFi.fileFormat = FileInfo.TIFF;
        lsmFi.pixelDepth = cz.VoxelSizeZ * 1000000;
        lsmFi.pixelHeight = cz.VoxelSizeY * 1000000;
        lsmFi.pixelWidth = cz.VoxelSizeX * 1000000;
        lsmFi.unit = "m";
        lsmFi.valueUnit = "m";
        lsmFi.nImages = 1;
        int datatype = (int) cz.IntensityDataType;
        switch (datatype) {
        case 1:
            lsmFi.fileType = FileInfo.GRAY8;
            break;
        case 2:
            lsmFi.fileType = FileInfo.GRAY16_UNSIGNED;
            break;
        case 5:
            lsmFi.fileType = FileInfo.GRAY32_FLOAT;
            break;
        default:
            lsmFi.fileType = FileInfo.GRAY8;
            break;
        }
        lsmFi.intelByteOrder = true;

        if (compression == 5) {
            lsmFi.compression = FileInfo.LZW;
            if (predictor == 2)
                lsmFi.compression = FileInfo.LZW_WITH_DIFFERENCING;
        } else
            lsmFi.compression = 0;
        lsmFi.stripOffsets = long2int(stripOffsets);
        lsmFi.stripLengths = long2int(stripLengths);
        return lsmFi;
    }

    private int[] long2int(long[] longArray) {
        int[] intArray = new int[longArray.length];
        for (int i = 0; i < longArray.length; i++)
            intArray[i] = (int) longArray[i];
        return intArray;
    }

    private ImagePlus[] readStack(RandomAccessStream stream, LsmFileInfo lsmFi,
            CZ_LSMInfo cz, ColorModel cm, FileOpener fop) {
        ImageStack[] stacks = new ImageStack[(int) cz.DimensionChannels];

        ImageDirectory imDir2 = (ImageDirectory) lsmFi.imageDirectories.get(0);
        for (int channelCount = 0; channelCount < (int) (cz.DimensionChannels); channelCount++) {
            stacks[channelCount] = new ImageStack((int) imDir2.TIF_IMAGEWIDTH,
                    (int) imDir2.TIF_IMAGELENGTH, cm);
        }
        imDir2 = null;
        for (int imageCounter = 0; imageCounter < lsmFi.imageDirectories.size(); imageCounter++) {
            ImageDirectory imDir = (ImageDirectory) lsmFi.imageDirectories
                    .get(imageCounter);
            if (imDir.TIF_NEWSUBFILETYPE == 0) {
                lsmFi.width = (int) imDir.TIF_IMAGEWIDTH;
                lsmFi.height = (int) imDir.TIF_IMAGELENGTH;
                Vector validOffsets = new Vector();
                for (int k = 0; k < imDir.TIF_STRIPOFFSETS.length; k++) {
                    validOffsets
                            .addElement(new Long(imDir.TIF_STRIPOFFSETS[k]));
                }
                lsmFi.stripOffsets = new int[imDir.TIF_STRIPOFFSETS.length];
                for (int i = 0; i < imDir.TIF_STRIPOFFSETS.length; i++)
                    lsmFi.stripOffsets[i] = (int) imDir.TIF_STRIPOFFSETS[i];
                ImageReader reader = new ImageReader(lsmFi);
                Object pixels;
                for (int channelCount = 0; channelCount < (int) (cz.DimensionChannels); channelCount++) {
                    int offset = ((Long) validOffsets.get(channelCount))
                            .intValue();
                    try {
                        stream.seek(offset);
                    } catch (IOException e) {

                        e.printStackTrace();
                    }
                    pixels = reader.readPixels((InputStream) stream);
                    stacks[channelCount].addSlice("", pixels);
                }
            }
        }
        ArrayList imps = new ArrayList();
        IJ.showProgress(1.0);
        for (int stackCount = 0; stackCount < (int) stacks.length; stackCount++) {
            ImagePlus imp = new ImagePlus(lsmFi.fileName, stacks[stackCount]);
            imp.setFileInfo(lsmFi);
            setCalibration(imp, lsmFi, fop);
            ImageProcessor ip = imp.getProcessor();
            if (ip.getMin() == ip.getMax())
                IJ.showProgress(1.0);
            imp.show();
            imp.setTitle(lsmFi.fileName + " Channel : "
                    + cz.channelNamesAndColors.ChannelNames[stackCount]);
            Color[] gc = new Color[2];
            gc[0] = new Color(0, 0, 0);
            int r = (int) (cz.channelNamesAndColors.Colors[stackCount] & 255);
            int g = (int) ((cz.channelNamesAndColors.Colors[stackCount] >> 8) & 255);
            int b = (int) ((cz.channelNamesAndColors.Colors[stackCount] >> 16) & 255);
            gc[1] = new Color(r, g, b);
            if (r == 0 && g == 0 && b == 0)
                gc[1] = Color.white;
            applyColors(imp, gc, 2);

            /*
             * imp.getWindow().addWindowListener( new
             * ImageFocusListener(masterModel, lsmFi, imp));
             */

            if (cz.IntensityDataType == 2) {
                double min = imp.getProcessor().getMin();
                double max = imp.getProcessor().getMax();
                imp.getProcessor().setMinAndMax(min, max);
            }
            imps.add(imp);
        }

        ImagePlus[] impsArr = new ImagePlus[imps.size()];
        for (int i = 0; i < imps.size(); i++)
            impsArr[i] = (ImagePlus) imps.get(i);
        return (impsArr);
    }

    private void setCalibration(ImagePlus imp, FileInfo fi, FileOpener fop) {
        Calibration cal = imp.getCalibration();
        if (fi.fileType == FileInfo.GRAY16_SIGNED) {
            if (IJ.debugMode)
                IJ.log("16-bit signed");
            double[] coeff = new double[2];
            coeff[0] = -32768.0;
            coeff[1] = 1.0;
            cal.setFunction(Calibration.STRAIGHT_LINE, coeff, "gray value");
        }

        // Properties props = decodeDescriptionString(fi);

        if (fi.pixelWidth > 0.0 && fi.unit != null) {
            cal.pixelWidth = fi.pixelWidth;
            cal.pixelHeight = fi.pixelHeight;
            cal.pixelDepth = fi.pixelDepth;
            cal.setUnit(fi.unit);
        }

        if (fi.valueUnit != null) {
            int f = fi.calibrationFunction;
            if ((f >= Calibration.STRAIGHT_LINE
                    && f <= Calibration.GAMMA_VARIATE && fi.coefficients != null)
                    || f == Calibration.UNCALIBRATED_OD)
                cal.setFunction(f, fi.coefficients, fi.valueUnit);
        }

        if (fi.frameInterval != 0.0)
            cal.frameInterval = fi.frameInterval;

    }

    /***************************************************************************
     * method : Apply_colors, applies color gradient; function taken out from *
     * my Lut_Panel plugin *
     **************************************************************************/

    private void applyColors(ImagePlus imp, Color[] gc, int i) {
        FileInfo fi = new FileInfo();
        fi.reds = new byte[256];
        fi.greens = new byte[256];
        fi.blues = new byte[256];
        fi.lutSize = 256;
        float nColorsfl = 256;
        float interval = 256;
        float iR = gc[0].getRed();
        float iG = gc[0].getGreen();
        float iB = gc[0].getBlue();
        float idR = gc[1].getRed() - gc[0].getRed();
        float idG = gc[1].getGreen() - gc[0].getGreen();
        float idB = gc[1].getBlue() - gc[0].getBlue();
        idR = (idR / interval);
        idG = (idG / interval);
        idB = (idB / interval);
        int a = 0;
        for (a = (int) (interval * 0); a < (int) (interval * (0) + interval); a++, iR += idR, iG += idG, iB += idB) {
            fi.reds[a] = (byte) (iR);
            fi.greens[a] = (byte) (iG);
            fi.blues[a] = (byte) (iB);
        }
        int b = (int) (interval * 0 + interval) - 1;
        fi.reds[b] = (byte) (gc[1].getRed());
        fi.greens[b] = (byte) (gc[1].getGreen());
        fi.blues[b] = (byte) (gc[1].getBlue());
        nColorsfl = 256;

        if (nColorsfl > 0) {
            if (nColorsfl < 256)
                interpolate(fi.reds, fi.greens, fi.blues, (int) nColorsfl);
            showLut(imp, fi, true);
            return;
        }
    }

    /***************************************************************************
     * method : interpolate taken out from the orignal ImageJ project from Wayne *
     * Rasband.No access to method so I had to add it manually. *
     **************************************************************************/

    private void interpolate(byte[] reds, byte[] greens, byte[] blues,
            int nColors) {
        byte[] r = new byte[nColors];
        byte[] g = new byte[nColors];
        byte[] b = new byte[nColors];
        System.arraycopy(reds, 0, r, 0, nColors);
        System.arraycopy(greens, 0, g, 0, nColors);
        System.arraycopy(blues, 0, b, 0, nColors);
        double scale = nColors / 256.0;
        int i1, i2;
        double fraction;
        for (int i = 0; i < 256; i++) {
            i1 = (int) (i * scale);
            i2 = i1 + 1;
            if (i2 == nColors)
                i2 = nColors - 1;
            fraction = i * scale - i1;
            reds[i] = (byte) ((1.0 - fraction) * (r[i1] & 255) + fraction
                    * (r[i2] & 255));
            greens[i] = (byte) ((1.0 - fraction) * (g[i1] & 255) + fraction
                    * (g[i2] & 255));
            blues[i] = (byte) ((1.0 - fraction) * (b[i1] & 255) + fraction
                    * (b[i2] & 255));
        }
    }

    /***************************************************************************
     * method : showLut, applies the new Lut on the actual image *
     **************************************************************************/

    private void showLut(ImagePlus imp, FileInfo fi, boolean showImage) {
        if (imp != null) {
            if (imp.getType() == ImagePlus.COLOR_RGB)
                IJ.error("Color tables cannot be assiged to RGB Images.");
            else {
                ImageProcessor ip = imp.getProcessor();
                ColorModel cm = new IndexColorModel(8, 256, fi.reds, fi.greens,
                        fi.blues);
                ip.setColorModel(cm);
                if (imp.getStackSize() > 1)
                    imp.getStack().setColorModel(cm);
                imp.updateAndDraw();
            }
        }
        /*
         * else imp.createImage(fi, showImage);
         */
    }

    private short swap(short x) {
        return (short) ((x << 8) | ((x >> 8) & 0xff));
    }

    private char swap(char x) {
        return (char) ((x << 8) | ((x >> 8) & 0xff));
    }

    private int swap(int x) {
        return (int) ((swap((short) x) << 16) | (swap((short) (x >> 16)) & 0xffff));
    }

    private long swap(long x) {
        return (long) (((long) swap((int) (x)) << 32) | ((long) swap((int) (x >> 32)) & 0xffffffffL));
    }

    private float swap(float x) {
        return Float.intBitsToFloat(swap(Float.floatToIntBits(x)));
    }

    private double swap(double x) {
        return Double.longBitsToDouble(swap(Double.doubleToLongBits(x)));
    }

    private class ImageDirectory {

        public long TIF_NEWSUBFILETYPE = 0;

        public long TIF_IMAGEWIDTH = 0;

        public long TIF_IMAGELENGTH = 0;

        public long TIF_BITSPERSAMPLE_LENGTH = 0;

        public long[] TIF_BITSPERSAMPLE_CHANNEL = new long[3];

        public long TIF_COMPRESSION = 0;

        public long TIF_PHOTOMETRICINTERPRETATION = 0;

        public long TIF_STRIPOFFSETS_LENGTH = 0;

        public long[] TIF_STRIPOFFSETS;

        public long TIF_SAMPLESPERPIXEL = 0;

        public long TIF_STRIPBYTECOUNTS_LENGTH = 0;

        public long[] TIF_STRIPBYTECOUNTS;

        public long TIF_PLANARCONFIGURATION = 0;

        public long TIF_PREDICTOR = 0;

        public long TIF_COLORMAP = 0;

        public long TIF_CZ_LSMINFO_OFFSET = 0; // OFFSET

        public CZ_LSMInfo TIF_CZ_LSMINFO; // STRUCT

        public long OFFSET_NEXT_DIRECTORY = 0;

    }

    private class LsmFileInfo extends FileInfo {
        public ArrayList imageDirectories = new ArrayList();
    }

    private class CZ_LSMInfo {

        public long DimensionChannels = 0;

        public long IntensityDataType = 0;

        public double VoxelSizeX = 0;

        public double VoxelSizeY = 0;

        public double VoxelSizeZ = 0;

        public int ScanType = 0;

        public long OffsetChannelColors = 0;

        public ChannelNamesAndColors channelNamesAndColors;

    }

    private class ChannelNamesAndColors {

        public long BlockSize = 0;

        public long NumberColors = 0;

        public long NumberNames = 0;

        public long ColorsOffset = 0;

        public long NamesOffset = 0;

        public long Mono = 0;

        public long[] Reserved;

        public long[] Colors;

        public String[] ChannelNames;

    }

}
